from pyspark.accumulators import Accumulator


class Ccf:
    class Iterate:
        @staticmethod
        def _map(key, value):
            return ((key, value), (value, key))

        @staticmethod
        def _reduce(key, values, accumulator: Accumulator = None):
            value_list = []
            to_return = []
            min_value = key
            for value in values:
                if value < min_value:
                    min_value = value
                value_list.append(value)
            print(value_list)
            if (min_value < key):
                to_return.append((key, min_value))
                for value in value_list:
                    if min_value != value:
                        to_return.append((value, min_value))
                        if accumulator:
                            accumulator.add(1)
            return to_return

        @staticmethod
        def map(x):
            return __class__._map(x[0], x[1])

        @staticmethod
        def reduce(x, accumulator: Accumulator = None):
            return __class__._reduce(x[0], x[1], accumulator)

    class IterateSecondarySorting(Iterate):
        @staticmethod
        def _reduce(key, values, accumulator: Accumulator = None):
            min_value = values[0]
            to_return = []
            if min_value < key:
                to_return.append((key, min_value))
                for value in values[1:]:
                    to_return.append((value, min_value))
                    if accumulator:
                        accumulator.add(1)
            return to_return

        @staticmethod
        def reduce(x, accumulator: Accumulator = None):
            return __class__._reduce(x[0], x[1], accumulator)

    class Dedup:
        @staticmethod
        def _map(key, value):
            return ((key, value), None)

        @staticmethod
        def _reduce(key, value):
            return (key[0], key[1])

        @staticmethod
        def map(x):
            return __class__._map(x[0], x[1])

        @staticmethod
        def reduce(x):
            return __class__._reduce(x[0], x[1])



import networkx as nx


class GraphGenerator:
    def __init__(self, nodes: int, edges: int):
        self.graph = self._generate_random_graph(nodes, edges)

    def generate_new_random_graph(self, nodes: int, edges: int):
        self.graph = self._generate_random_graph(nodes, edges)

    def _generate_random_graph(self, nodes: int, edges: int):
        return nx.gnm_random_graph(nodes, edges)

    def remove_null_degres_nodes(self):
        copy = self.graph.copy()
        self.graph = nx.subgraph_view(
            copy,
            filter_node=lambda node: nx.degree(copy, node) != 0,
        )

    def save(self, path):
        with open(path, 'wb') as f:
            nx.write_edgelist(self.graph, f)

    def draw(self):
        nx.draw(self.graph, with_labels=True)

    @staticmethod
    def generate_random_graph(nodes, edges):
        graph = GraphGenerator(nodes, edges)
        graph.remove_null_degres_nodes()
        return graph

    @property
    def edges(self):
        return self.graph.edges

    @property
    def nodes(self):
        return self.graph.nodes



from networkx.classes.graph import Graph
from pyspark import SparkContext
from ccf_spark.graph_generator import GraphGenerator
from ccf_spark.ccf import Ccf

CCF_DEDUP = Ccf.Dedup
CCF_ITERATE = Ccf.Iterate
CCF_ITERATE_SECONDARY_SORTING = Ccf.IterateSecondarySorting


class CcfSpark:
    def __init__(self,
                 sc: SparkContext,
                 secondary_sorting: bool = False,
                 graph: Graph = None,
                 file_path: str = None):
        self.sc = sc
        self.iterator = CCF_ITERATE_SECONDARY_SORTING if secondary_sorting else CCF_ITERATE
        self.secondary_sorting = secondary_sorting
        if graph:
            self.graph = sc.parallelize(self.graph.edges)
        elif file_path:
            # File line format expected : <int> <int>
            self.graph = sc.textFile(file_path).map(
                lambda x: tuple(map(int,
                                    x.split('  ')[:2])))
        else:
            self.graph = sc.parallelize(
                GraphGenerator.generate_random_graph(500, 350).edges)

    def iterate(self):
        accumulator = self.sc.accumulator(0)
        iterator = self.iterator  # To avoid SPARK-5063 error.
        self.graph = self.graph.flatMap(iterator.map).groupByKey()
        if self.secondary_sorting:
            self.graph = self.graph.map(lambda x: (x[0], sorted(x[1])))
        self.graph = self.graph.flatMap(
            lambda x, accumulator=accumulator: iterator.reduce(x, accumulator
                                                               )).sortByKey()
        self.graph = self.graph.map(CCF_DEDUP.map).groupByKey()
        self.graph = self.graph.map(CCF_DEDUP.reduce)
        return accumulator.value

    def iterate_all(self):
        while True:
            new_pairs = self.iterate()
            if not new_pairs:
                break

    def print(self):
        return self.graph.collect()





from distutils.core import setup
from os import path

# dir = path.abspath(path.dirname(__file__))
# with open(path.join(dir, 'README.md'), encoding='utf-8') as f:
#     long_description = f.read()

setup(
    name='ccf_spark',
    version='0.1.1',
    description='A little CCF implementation in a Spark context with networkx',
    # long_description=long_description,
    # long_description_content_type=text/markdown,
    author='Théo Chennebault',
    author_email='theo.chennebault@le-cab-politique.fr',
    url='https://github.com/ErnestBidouille/ccf-spark',
    packages=[
        'ccf_spark',
    ],
    install_requires=['networkx==2.5', 'pyspark==3.0.1'])
